import re
from collections import defaultdict

import numpy as np
import pandas as pd

english_words = set()
with open("src/finney/data/words.txt", "r") as f:
    for line in f.readlines():
        english_words.add(line.strip().casefold())
temp = "|".join(map(re.escape, english_words))
english_pattern = re.compile(rf"\b(?:{temp})\b")

keywords = set()
with open("src/finney/data/keywords.txt", "r") as f:  # taken from https://github.com/e3b0c442/keywords?tab=readme-ov-file
    for line in f.readlines():  # and from https://www.ibm.com/docs/en/i/7.6.0?topic=extensions-standard-c-library-functions-table-by-name
        keywords.add(line.strip().casefold())
temp = "|".join(map(re.escape, keywords))
keyword_pattern = re.compile(rf"\b(?:{temp})\b")

extensions = set()  # taken from https://gist.github.com/securifera/e7eed730cbe1ce43d0c29d7cd2d582f4
with open("src/finney/data/extensions.txt", "r") as f:
    for line in f.readlines():
        extensions.add(line.strip().casefold())
temp = "|".join(map(re.escape, extensions))
file_type_pattern = re.compile(rf"(?:{temp})$")

domains = set()  # taken from https://github.com/datasets/top-level-domain-names/blob/main/data/top-level-domain-names.csv?plain=1
with open("src/finney/data/domains.txt", "r") as f:
    for line in f.readlines():
        domains.add(line.strip().casefold())
temp = "|".join(map(re.escape, domains))
url_pattern = re.compile(rf"(?:{temp})\b")

key_distances = pd.read_csv("src/finney/data/bigrams.csv", index_col=0).to_numpy()
key_index = {ch: i for i, ch in enumerate("!@#$%^&*()_+1234567890-=qwertyuiop[]{}asdfghjkl;'\\:\"|~zxcvbnm,./<>?)}")}

character_type_map = defaultdict(int)
for c in "abcdefghijklmnopqrstuvwxyz":
    character_type_map[c] = 1
for c in "0123456789":
    character_type_map[c] = 2


def extract_bigrams(word: str) -> list[tuple[str, str]]:
    if len(word) < 2:
        return []
    word = word.lower().strip()
    return list(zip(word, word[1:]))


def avg_key_distance(bigrams: list[tuple[str, str]]) -> np.float32:
    total_distance = np.float32(0)
    if not bigrams:
        return total_distance
    for bigram in bigrams:
        c1, c2 = bigram
        if c1 not in key_index or c2 not in key_index:
            continue
        total_distance += key_distances[key_index[c1]][key_index[c2]]  # type: ignore
    return total_distance / len(bigrams)


def count_type_switches(bigrams: list[tuple[str, str]]) -> int:
    switch_count = 0
    for bigram in bigrams:
        c1, c2 = bigram
        switch_count += character_type_map[c1] != character_type_map[c2]
    return switch_count


def extract_trigrams(word: str) -> list[tuple[str, str, str]]:
    if len(word) < 3:
        return []
    word = word.lower().strip()
    return list(zip(word, word[1:], word[2:]))


def has_consecutive_sequence(trigrams: list[tuple[str, str, str]]) -> bool:
    for trigram in trigrams:
        c1, c2, c3 = trigram
        if abs(ord(c1) - ord(c2)) == 1 and ord(c1) - ord(c2) == ord(c2) - ord(c3):
            return True
    return False


def get_features(df: pd.DataFrame) -> pd.DataFrame:
    out = pd.DataFrame()

    # contains special character
    out["special"] = df["text"].str.contains("[!@#\$\^\&\*\(\)_\+\[\]'\"\;\/\,\>\<\\\|\{\}\?\.]", regex=True)

    # ends with special character
    out["special_end"] = df["text"].str.contains("[^a-zA-Z0-9]$", regex=True)

    # starts with uppercase
    out["upper_start"] = df["text"].str.contains("^[A-Z]", regex=True)

    # contains a space
    # df["space"] = df["text"].str.contains(" ")

    # is entirely hexadecimal, with at least one non-digit character
    out["all_hexa"] = df["text"].str.contains("^[0-9A-Fa-f]*[A-Fa-f][0-9A-Fa-f]*$", regex=True)

    # contains escaped byte (e.g. for urls)
    out["byte"] = df["text"].str.contains(r"\\{1,2}\w{3,4}(?!\w)", regex=True)

    # starts with one of . .. ./ ../ like a relative path path
    out["path_sub"] = df["text"].str.contains(r"(?:^/\./)|(?:^\./)", regex=True)
    out["path_relative"] = df["text"].str.contains(r"(?:^/\.\./)|(?:^\.\./)", regex=True)

    # is of common password format: [letters][numbers][symbol] or [letters][symbol][number]
    out["letter_number_symbol"] = df["text"].str.contains(r"^[A-Za-z]+[0-9]+\W?$", regex=True)
    out["letter_symbol_number"] = df["text"].str.contains(r"^[A-Za-z]+\W?[0-9]+$", regex=True)

    # contains a number of a recent or upcoming year (1900-2100), or a date formatted YYYY-MM-DD
    out["year"] = df["text"].str.contains(r"\b(?:19[0-9]{2}|20[0-9]{2}|2100)\b", regex=True)
    out["date"] = df["text"].str.contains(r"\d{4}-\d{2}-\d{2}", regex=True)

    # contains an xml/html tag (e.g. <div>)
    out["xml"] = df["text"].str.contains(r"<.{1,3}>", regex=True)

    # contains specific special characters that are kinda common in code
    out["period"] = df["text"].str.contains("\.")
    out["double_colon"] = df["text"].str.contains("::")
    out["question"] = df["text"].str.contains("\?")
    out["percent"] = df["text"].str.contains("%")
    out["arrow"] = df["text"].str.contains("->")
    out["dunder"] = df["text"].str.contains("__")
    out["double_equal"] = df["text"].str.contains("==")
    out["triple_equal"] = df["text"].str.contains("===")
    out["double_slash"] = df["text"].str.contains("//")
    out["backslash"] = df["text"].str.contains("\\\\")
    out["double_backslash"] = df["text"].str.contains("\\\\\\\\")
    out["newline"] = df["text"].str.contains(r"\\n", regex=True)

    # contains an equal number of opening and closting parentheses of various types
    out["balanced_parentheses"] = (
            (df["text"].str.count(r"\(") == df["text"].str.count(r"\)")) &
            (df["text"].str.count(r"\(") > 0)
    )
    out["balanced_parentheses_square"] = (
            (df["text"].str.count(r"\[") == df["text"].str.count(r"\]")) &
            (df["text"].str.count(r"\[") > 0)
    )
    out["balanced_parentheses_curl"] = (
            (df["text"].str.count(r"\{") == df["text"].str.count(r"\}")) &
            (df["text"].str.count(r"\{") > 0)
    )

    # longest sequence of uppercase letters, vowels, consonants, and hexadecimal characters
    out["longest_upper"] = df["text"].apply(
        lambda x: max((len(m) for m in re.findall(r"[A-Z]+", str(x))), default=0)
    )
    out["longest_vowels"] = df["text"].apply(
        lambda x: max((len(m) for m in re.findall(r"[aeiouAEIOU]+", str(x))), default=0)
    )
    out["longest_cons"] = df["text"].apply(
        lambda x: max((len(m) for m in re.findall(r"[bcdfghjklmnpqrstvxzBCDFGHJKLMNPQRSTVXZ]+", str(x))), default=0)
    )
    out["longest_hexa"] = df["text"].apply(
        lambda x: max((len(m) for m in re.findall(r"[0-9A-Fa-f]*[A-Fa-f][0-9A-Fa-f]*", str(x))), default=0)
    )

    # fraction of the string that's digits, letters, and other
    out["digit_fraction"] = df["text"].apply(
        lambda x: sum(c.isdigit() for c in str(x)) / len(str(x))
    )
    out["vowel_fraction"] = df["text"].apply(
        lambda x: sum(c in set("aeiouAEIOU") for c in str(x)) / len(str(x))
    )
    out["nonword_fraction"] = df["text"].apply(
        lambda x: len(re.findall(r"\W", str(x))) / len(str(x))
    )

    # number of character sequences of length divisible by 4
    out["word_length_mod_4"] = df["text"].apply(
        lambda x: sum(1 for m in re.findall(r"\w+", str(x)) if len(m) % 4 == 0)
    )

    # whether the string is in a common programming style convention
    snake = r'^[A-Za-z]+(?:_[A-Za-z_]+)+$'
    pascal = r'^[A-Z][a-z]+(?:[A-Z][a-z]+)+$'  # UpperCamel
    camel = r'^[a-z]+(?:[A-Z][a-z]+)+$'  # lowerCamel
    kebab = r'^[A-Za-z]+(?:-[A-Za-z-]+)+$'  # kebab-like (at least one "-")
    allcaps = r'^[A-Z]+$'  # ALLCAPS

    conds = [
        df["text"].str.fullmatch(snake, na=False),
        df["text"].str.fullmatch(pascal, na=False),
        df["text"].str.fullmatch(camel, na=False),
        df["text"].str.fullmatch(kebab, na=False),
        df["text"].str.fullmatch(allcaps, na=False),
    ]
    out["format"] = np.select(conds, [1, 2, 3, 4, 5], default=0).astype(int)

    # whether the string contains an english word, and how many
    out["word"] = df["text"].astype(str).str.contains(english_pattern, na=False, regex=True)
    out["word_count"] = df["text"].astype(str).str.findall(english_pattern).str.len()

    # whether the string contains a programming keyword, and how many
    out["keyword"] = df["text"].astype(str).str.contains(keyword_pattern, na=False, regex=True)
    out["keyword_count"] = df["text"].astype(str).str.findall(keyword_pattern).str.len()

    # is a likely url
    out["likely_url"] = df["text"].astype(str).str.contains(url_pattern, na=False)

    # ends with known file suffix (e.g. .exe or .py)
    out["file_suffix"] = df["text"].astype(str).str.contains(url_pattern, na=False)

    # string length
    out["string_length"] = df["text"].astype(str).str.len()

    # split the word into bigrams (e.g. "bigram" -> [bi, ig, gr, ra, am])
    bigrams = df.text.map(lambda word: extract_bigrams(word))
    out["key_distances"] = list(map(avg_key_distance, bigrams))
    out["type_switches"] = list(map(count_type_switches, bigrams))

    # split the word into trigrams (e.g. "trigram" -> [tri, rig, igr, gra, ram])
    trigrams = df.text.map(lambda word: extract_trigrams(word))
    out["consecutive_sequence"] = list(map(has_consecutive_sequence, trigrams))

    # return tuple(features)
    return out



import os
import pickle
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Sequence

import click
from rich.console import Console
from rich.table import Table
from rich import box
import yaml

from finney import search
from finney.domain_objects import Match, IgnoreConfig

root = ".finney"
config_path = f"{root}/config"
last_matches_path = f"{root}/matches"

if not os.path.exists(root):
    os.mkdir(root)


class ENTRY_TYPE(Enum):
    STRINGS = "STRINGS"
    FILES = "FILES"
    DIRS = "DIRS"
    TYPES = "SUFFIXES"  # finney: ignore


class MODE(Enum):
    ADD = "ADD"
    SUBTRACT = "SUBTRACT"


def _get_recursive_paths(paths: list[str]) -> list[str]:
    files = []
    for dir_path in paths:
        path = Path(dir_path)
        if path.is_dir():
            files.extend(str(p) for p in path.rglob('*') if p.is_file())
    return files


def _load_ignore_config() -> IgnoreConfig:
    if not os.path.exists(config_path):
        return IgnoreConfig(dirs=[], files=[], types=[], strings=[])

    with open(config_path, "r") as f:
        config = yaml.safe_load(f) or {}

    config = config.get("ignore", {})
    return IgnoreConfig(
        dirs=config.get("dirs") or [],
        files=config.get("files") or [],
        types=config.get("types") or [],
        strings=config.get("strings") or [],
    )


def _edit_ignore_entries(
        entry_type: ENTRY_TYPE, mode: MODE, values: Sequence[str]
) -> None:
    prev_config = _load_ignore_config()

    if mode == MODE.ADD and entry_type == ENTRY_TYPE.STRINGS:
        last_run_matches = _load_last_matches()
        temp = set(values)
        for value in values:
            if value not in last_run_matches:
                if not click.confirm(f"String {value} was not found in the last run. Are you sure you want to ignore it?", default=True, prompt_suffix="\n>>> "):
                    temp.remove(value)
        values = list(temp)


    added_config = IgnoreConfig(
        dirs=values if entry_type == ENTRY_TYPE.DIRS else [],
        files=values if entry_type == ENTRY_TYPE.FILES else [],
        types=values if entry_type == ENTRY_TYPE.TYPES else [],
        strings=values if entry_type == ENTRY_TYPE.STRINGS else [],
    )
    combined = (
        prev_config + added_config if mode == MODE.ADD else prev_config - added_config
    )
    with open(config_path, "w+") as f:
        yaml.safe_dump(
            {"ignore": combined.to_dict()},
            f,
            indent=4,
            default_flow_style=False,
            sort_keys=False,
        )
    entry_str = "entry" if len(values) == 1 else "entries"
    if mode == MODE.ADD:
        print(f"Added {len(values)} {entry_str} to ignore list")
    else:
        print(f"Removed {len(values)} {entry_str} from ignore list")


def _select_entry_type(
        strings: bool, files: bool, dirs: bool, types: bool
) -> ENTRY_TYPE:
    if sum([strings, files, dirs, types]) > 1:
        raise click.UsageError("Options -s, -f, -d, -t, and -i are mutually exclusive")

    elif types:
        return ENTRY_TYPE.TYPES

    elif files:
        return ENTRY_TYPE.FILES

    elif dirs:
        return ENTRY_TYPE.DIRS

    elif strings:
        return ENTRY_TYPE.STRINGS

    else:
        return ENTRY_TYPE.STRINGS


def _save_last_matches(matches: Sequence[Match]) -> None:
    with open(last_matches_path, "wb+") as f:
        pickle.dump(matches, f)


def _load_last_matches() -> Sequence[str]:
    if not os.path.exists(last_matches_path):
        return []
    with open(last_matches_path, "rb") as f:
        matches: list[Match] = pickle.load(f)
    return [m.sha for m in matches]


@click.group(help="Scan your code repositories for hardcoded passwords and secrets")
def cli():
    pass


def _matches_by_file(matches: Sequence[Match]) -> dict[str, list[Match]]:
    out = defaultdict(list)
    for match in matches:
        out[str(match.path)].append(match)
    return out


def _print_match_group(matches: list[Match]) -> None:
    matches.sort(key=lambda x: x.line)
    table = Table(box=box.MINIMAL)
    print(f"In file: {matches[0].path}")
    table.add_column("Line", justify="right")
    table.add_column("Suspected Secret", justify="left")

    for m in matches:
        table.add_row(str(m.line), m.match)

    console = Console()
    console.print(table)


def _pretty_print(matches: Sequence[Match]) -> None:
    match_groups = _matches_by_file(matches)
    matches_str = f"{len(matches)} suspected {'secrets' if len(matches) > 1 else 'secret'}"
    files_str = f"{len(match_groups)} {'files' if len(match_groups) > 1 else 'file'}"
    click.secho(
        f"Found {matches_str} in {files_str}:\n"
    )
    for path, group in sorted(match_groups.items(), key=lambda item: item[1][0].path):
        _print_match_group(group)
        print()
    print("""If these are real secrets, consider removing them from your code before committing.
If they aren't, you can mark them as safe in the following ways:
- Ignore specific strings explicity:
    finney ignore [STRINGS...]
    
- Ignore entire files:
    finney ignore -f [FILE_NAME...]
    
- Ignore specific lines by appending to them
    # finney: ignore""")


@cli.command(help="Run Finney on the given files")
@click.argument("paths", nargs=-1)
@click.option("-r", "recursive", is_flag=True, default=False, help="Recursively search the given paths")
def run(paths, recursive):
    ignored = _load_ignore_config()
    if recursive:
        paths = _get_recursive_paths(paths)
    matches: Sequence[Match] = search.scan_files(paths, ignored)

    if matches:
        _pretty_print(matches)
        _save_last_matches(matches)
        exit(1)
    print("Finney didn't find any suspected secrets :D")


@cli.command("ignore", help="Defined values that can be safely ignored")
@click.option("-s", "strings", is_flag=True, help="Define specific strings as safe (default)")
@click.option("-f", "files", is_flag=True, help="Define files that Finney won't scan")
@click.option("-d", "dirs", is_flag=True, help="Define directories that Finney won't scan")
@click.option("-t", "types", is_flag=True, help="Define file types (.exe, .jar, ...) that Finney won't scan")
@click.argument("values", nargs=-1)
def ignore(strings, files, dirs, types, values):
    entry_type = _select_entry_type(strings, files, dirs, types)
    _edit_ignore_entries(entry_type, mode=MODE.ADD, values=list(values))


@cli.command("unignore", help="Remove values from the ignore list.\nSee `finney ignore` for details`")
@click.option("-s", "strings", is_flag=True, help="Remove strings from the ignored list (default)")
@click.option("-f", "files", is_flag=True, help="Remove files from the ignored list")
@click.option("-d", "dirs", is_flag=True, help="Remove directories from the ignore list")
@click.option("-t", "types", is_flag=True, help="Remove file types from the ignore list")
@click.argument("values", nargs=-1)
def unignore(strings, files, dirs, types, values):
    entry_type = _select_entry_type(strings, files, dirs, types)
    _edit_ignore_entries(entry_type, mode=MODE.SUBTRACT, values=list(values))


@cli.command("list", help="Print the current ignore configuration")
def _list():
    config = _load_ignore_config()
    config.print()


if __name__ == "__main__":
    cli()



import pickle
import time
from itertools import combinations_with_replacement

import pandas as pd
from sklearn.model_selection import GridSearchCV, train_test_split
import xgboost as xgb

from finney.models.features import get_features

alphabet = list("abcdefghijklmnopqrstuvwxyz")
short_words = alphabet + ["".join(x) for x in combinations_with_replacement(alphabet, 2)]

snippet_words_df = list(pd.read_csv("src/finney/data/context_words.csv"))


def run_grid_search():
    samples = 10_000
    param_grid = {
        "eta": [0.01, 0.1, 0.2, 0.3, 0.4, 0.5],
        "max_depth": [5, 10, 15, 20, 25],
        "n_estimators": [100, 500, 1000, 2000],
    }

    df = pd.read_csv(
        "/finney/data/PassFInder_Password_Dataset/password_test.csv",
        header=None,
        names=["text", "label"],
    ).sample(samples)
    texts = df["text"].astype(str).tolist() + short_words + snippet_words_df
    labels = df["label"].astype(int).tolist() + [0 for _ in short_words] + [0 for _ in snippet_words_df]
    # labels = [y > 0 for y in labels]

    texts = pd.DataFrame(texts, columns=["text"])
    word_features = get_features(texts)

    texts_train, texts_test, X_train, X_test, y_train, y_test = train_test_split(
        texts, word_features, labels, test_size=0.2)

    clf = xgb.XGBClassifier(random_state=42)

    # def confusion_matrix_scorer(clf, X, y):
    #     y_pred = clf.predict(X)
    #     cm = confusion_matrix(y, y_pred)
    #     return {'tn1': cm[0, 0], 'fp1': cm[0, 1], 'fp2': cm[0,2],
    #             'fn1': cm[1, 0], 'tp1': cm[1, 1], 'tp2': cm[1,2],
    #             'fn2': cm[2, 0], 'tp3': cm[2, 1], 'tp4': cm[2,2]}

    grid_search = GridSearchCV(estimator=clf, param_grid=param_grid, cv=3, n_jobs=-1, verbose=2,
                                         scoring="f1", return_train_score=True).fit(X_train, y_train)
    # grid_search_recall = GridSearchCV(estimator=clf, param_grid=param_grid, cv=3, n_jobs=-1, verbose=2,
    #                                   scoring=confusion_matrix_scorer, return_train_score=True).fit(X_train, y_train)

    with open("src/finney/models/grid_search.pkl", "wb") as f:
        pickle.dump(grid_search, f)
    time.sleep(0.5)

    print("done?")

if __name__ == "__main__":
    run_grid_search()



from dataclasses import dataclass
from hashlib import sha256
from pathlib import Path
from typing import List


def _sub(l1, l2):
    return [x for x in l1 if x not in l2]


@dataclass
class Match:
    path: Path
    match: str
    line: int = 0

    def __str__(self):
        return f"{self.path}:{self.line:0>3}: '{self.match}'"

    @property
    def sha(self):
        return sha256(self.match.encode("utf-8")).hexdigest()

    @property
    def file(self):
        return self.path.name

    def __len__(self):
        return len(self.match) + 4

    def render(self, length):
        return "." * (length - len(str(self.match))) + " '" + self.match + "'"

    def __eq__(self, other):
        return self.path == other.path and self.match == other.match

    def __hash__(self):
        return hash(self.path.name + self.match)


@dataclass
class IgnoreConfig:
    dirs: List[Path]
    files: List[Path]
    types: List[str]
    strings: List[bytes]

    def __add__(self, other):
        if not isinstance(other, IgnoreConfig):
            raise TypeError
        return IgnoreConfig(
            self.dirs + other.dirs,
            self.files + other.files,
            self.types + other.types,
            self.strings + other.strings,
        )

    def __sub__(self, other):
        if not isinstance(other, IgnoreConfig):
            raise TypeError
        return IgnoreConfig(
            _sub(self.dirs, other.dirs),
            _sub(self.files, other.files),
            _sub(self.types, other.types),
            _sub(self.strings, other.strings),
        )

    def to_dict(self):
        return self.__dict__

    def print(self):
        if self.dirs:
            print("Directories:")
            for dir in self.dirs:
                print(f" - {dir}")
            print()

        if self.files:
            print("Files:")
            for f in self.files:
                print(f" - {f}")
            print()

        if self.types:
            print("File Types:")
            for t in self.types:
                print(f" - {t}")
            print()

        if self.strings:
            print("Strings:")
            for s in self.strings:
                print(f" - {s}")


import dataclasses
import pickle
import re
import time
from datetime import datetime
from itertools import combinations_with_replacement
from typing import Self

import numpy as np
import pandas as pd

from finney.models.features import get_features

candidate_pattern = r"""(["'`])[a-zA-Z0-9&*!?.\-_#%@^&$"'`{} ()\[\]]{6,30}\1"""


def extract_candidates_from_file(path) -> pd.DataFrame:
    with open(path, "r") as f:
        candidates = []
        for line in f.readlines():
            if match := re.search(candidate_pattern, line):
                candidates.append(match.group()[1:-1])
    return pd.DataFrame(candidates, columns=["text"])


alphabet = list("abcdefghijklmnopqrstuvwxyz")
short_words = alphabet + ["".join(x) for x in combinations_with_replacement(alphabet, 2)]

snippet_words_df = list(pd.read_csv("src/finney/data/context_words.csv"))


@dataclasses.dataclass
class Score:
    eta: float
    n_estimators: int
    max_depth: int
    samples: int
    train_accuracy: float
    train_precision: float
    train_recall: float
    train_f1: float
    test_accuracy: float
    test_precision: float
    test_recall: float
    test_f1: float

    @staticmethod
    def avg(scores: list['Self']):
        return Score(
            eta=scores[0].eta,
            n_estimators=scores[0].n_estimators,
            max_depth=scores[0].max_depth,
            samples=scores[0].samples,
            train_accuracy=sum([sc.train_accuracy for sc in scores]) / len(scores),
            train_precision=sum([sc.train_precision for sc in scores]) / len(scores),
            train_recall=sum([sc.train_recall for sc in scores]) / len(scores),
            train_f1=sum([sc.train_f1 for sc in scores]) / len(scores),
            test_accuracy=sum([sc.test_accuracy for sc in scores]) / len(scores),
            test_precision=sum([sc.test_precision for sc in scores]) / len(scores),
            test_recall=sum([sc.test_recall for sc in scores]) / len(scores),
            test_f1=sum([sc.test_f1 for sc in scores]) / len(scores),
        )

    def __str__(self):
        return f"{self.eta},{self.n_estimators},{self.max_depth},{self.samples},{self.train_accuracy},{self.test_accuracy},{self.train_precision},{self.test_precision},{self.train_recall},{self.test_recall},{self.train_f1},{self.test_f1}"


def make_tree(word_features: pd.DataFrame, samples: int, eta: float, max_depth: int, n_estimators: int,
              save: bool = True):
    import xgboost as xgb
    from sklearn.metrics import (
        accuracy_score,
        precision_score,
        recall_score,
        f1_score,
        confusion_matrix
    )
    from sklearn.model_selection import train_test_split

    texts_train, texts_test, X_train, X_test, y_train, y_test = train_test_split(
        texts, word_features, labels, test_size=0.2, random_state=42
    )

    clf = xgb.XGBClassifier(max_depth=max_depth, n_estimators=n_estimators, eta=eta)

    clf = clf.fit(np.array(X_train), np.array(y_train))  # train the model

    y_train = np.array(y_train)
    train_preds = clf.predict(X_train)
    train_accuracy = accuracy_score(y_train > 0, train_preds > 0)
    train_precision = precision_score(y_train > 0, train_preds > 0, average="macro")
    train_recall = recall_score(y_train > 0, train_preds > 0, average="macro")
    train_f1 = f1_score(y_train > 0, train_preds > 0, average="macro")

    confusion = confusion_matrix(y_train, train_preds)
    print(confusion)


    y_test = np.array(y_test)
    test_preds = clf.predict(X_test)
    test_accuracy = accuracy_score(y_test > 0, test_preds > 0)
    test_precision = precision_score(y_test > 0, test_preds > 0, average="macro")
    test_recall = recall_score(y_test > 0, test_preds > 0, average="macro")
    test_f1 = f1_score(y_test > 0, test_preds > 0, average="macro")

    confusion = confusion_matrix(y_test, test_preds)
    print(confusion)

    score = Score(
        eta=eta,
        max_depth=max_depth,
        n_estimators=n_estimators,
        samples=samples,
        train_accuracy=train_accuracy,
        train_precision=train_precision,
        train_recall=train_recall,
        train_f1=train_f1,
        test_accuracy=test_accuracy,
        test_precision=test_precision,
        test_recall=test_recall,
        test_f1=test_f1,
    )
    # print("Performance metrics:")
    # print(f"  {accuracy}")
    # print(f"  {precision}")
    # print(f"  {recall}")
    # print(f"  {f1}")

    if save:
        with open("src/finney/models/tree.pkl", "wb") as f:
            pickle.dump(clf, f)
            time.sleep(0.5)

    # df = pd.DataFrame(texts_test)
    # df["y_true"] = y_test
    # df["y_pred"] = preds.astype(bool)
    # errors = df[df["y_true"] != df["y_pred"]]

    return score


def predict(words):
    with open("src/finney/models/tree.pkl", "rb") as f:
        tree = pickle.load(f)
    words = pd.DataFrame(words)
    word_features = get_features(words)
    res = tree.predict_proba(word_features)
    return res


def clean_results(pred_weights, threshold):
    cond1 = pred_weights[:, 0] < threshold
    cond2 = pred_weights[:, 0] != pred_weights.max(axis=1)
    mask = cond1 & cond2
    indices = np.where(mask)[0].tolist()
    return indices


def scan(path, threshold=0.2):
    # try:
    candidates = extract_candidates_from_file(path)
    # except:
    # return []
    if not len(candidates.index):
        return []
    pred_weights = predict(candidates)
    results = clean_results(pred_weights, threshold)

    return [candidates.loc[i].text for i, guess in enumerate(results) if guess]


if __name__ == "__main__":
    samples = 5_000_000

    now = datetime.now()
    print(f"[{now.hour:0>2}:{now.minute:0>2}:{now.second:0>2}] Starting to compute features")

    # df = pd.read_csv(
    #     "/Users/danylewin/thingies/university/CS Workshop/Finney/data/PassFInder_Password_Dataset/password_test.csv",
    #     header=None,
    #     names=["text", "label"],
    # ).sample(int(samples*1.01)).dropna().sample(samples)

    df = pd.read_csv(
        "/finney/data/PassFInder_Password_Dataset/password_test.csv",
        header=None,
        names=["text", "label"],
    ).dropna()

    texts = df["text"].astype(str).tolist() + short_words + snippet_words_df
    labels = df["label"].astype(int).tolist() + [0 for _ in short_words] + [0 for _ in snippet_words_df]
    # labels = [y > 0 for y in labels]

    texts = pd.DataFrame(texts, columns=["text"])
    word_features = get_features(texts)

    with open("src/finney/models/features.pkl", "wb") as f:
        pickle.dump(word_features, f)
        time.sleep(0.5)

    now = datetime.now()
    print(f"[{now.hour}:{now.minute}:{now.second}] Finished computing features, starting run")

    with open("scores.csv", "a") as f:
        f.write(
            "eta,n_estimators,max_depth,samples,train_accuracy,train_precision,train_recall,train_f1,test_accuracy,test_precision,test_recall,test_f1\n")

    for eta in [0.05]:
        for max_depth in [15]:
            for n_estimators in [1000]:
                scores = []
                print(f"Running for {n_estimators=} {max_depth=} {eta=} {samples=}", end=" ")
                start = datetime.now()
                for i in range(1):
                    print(i + 1, end=" ")
                    scores.append(make_tree(
                        word_features=word_features,
                        eta=eta,
                        max_depth=max_depth,
                        n_estimators=n_estimators,
                        samples=samples,
                        save=True,
                    ))
                    avg = Score.avg(scores)
                    with open("scores.csv", "a") as f:
                        f.write(str(avg) + "\n")
                print(f"took: {datetime.now() - start}")



from pathlib import Path
from typing import Sequence

import click

from finney.domain_objects import Match, IgnoreConfig
from finney.models import intrinsic, decision_tree


def should_scan(file: Path, ignored: IgnoreConfig) -> bool:
    if file.suffix in ignored.types:
        return False
    if file.name in ignored.files:
        return False
    for dir in ignored.dirs:
        if dir in file.parts:
            return False
    return True


def find_lines(matches: list[Match]) -> list[Match]:
    out = []
    for match in matches:
        with open(match.path, "r") as f:
            for i, line in enumerate(f.readlines(), start=1):
                if match.match in line:
                    if "finney: ignore" not in line.casefold():
                        out.append(Match(match.path, match.match, i))
                    break
            else:
                raise ValueError(f"Expected to find suspected secret '{match.match}' in file '{match.path}'")
    return out


keywords = set()
with open("src/finney/data/keywords.txt", "r") as f:  # taken from https://github.com/e3b0c442/keywords?tab=readme-ov-file
    for line in f.readlines():
        keywords.add(line.strip().casefold())


def clean_matches(matches: list[Match]) -> list[Match]:
    matches = [m for m in matches if m.match.casefold() not in keywords]
    return list(set(matches))


def make_paths_relative(paths: list[Path]) -> list[Path]:
    out = []
    for p in paths:
        p = p.expanduser()
        if not p.is_absolute():
            p = Path.cwd() / p
        out.append(p.resolve(strict=False))
    return out

def scan_files(paths: Sequence[str], ignored: IgnoreConfig) -> list[Match]:
    files = [Path(f) for f in paths]
    matches = []
    hide_bar = len(paths) < 10
    with click.progressbar(files, label="Scanning files", hidden=hide_bar, show_pos=True) as bar:
        for file in bar:
            if not should_scan(file, ignored):
                continue
            try:
                res = intrinsic.scan(file, ignored)
                matches.extend([Match(file, s) for s in res])

                res = decision_tree.scan(file)
                matches.extend([Match(file, s) for s in res])
            except Exception as e:
                print(f"Failed to scan {file}")
                raise e
    if len(matches) > 1000:
        print("Collecting Results...")
    matches = find_lines(clean_matches(matches))
    return matches
